"""English localization."""

from collections import defaultdict
from typing import Any, Dict, Generator, List

import raesl.doc.lines as lns
from raesl.doc.locales import hookimpl, pm
from ragraph.graph import Graph
from ragraph.node import Node

LineGen = Generator[str, None, None]

translations = {"world": "the world"}


@hookimpl
def gettext(key: str):
    """Get translated string."""
    if key in translations:
        return translations[key]
    return key


_ = gettext


@hookimpl
def linguistic_enumeration(items: List[str]) -> str:
    n = len(items)
    if n <= 2:
        return " and ".join(items)

    return ", ".join(items[:-1]) + ", and " + items[-1]


@hookimpl
def linguistic_options(items: List[str]) -> str:
    return " or ".join(items)


# Node based hook implementations
def subclause_line(s: Dict[str, Any]):
    """Yield subclause line."""

    yield " or ".join([design_rule_line(r=r) for r in s["body"]])


def is_number(v: Any) -> bool:
    try:
        float(v)
        return True
    except ValueError:
        return False


def design_rule_line(r: Dict[str, str]):
    s = r["subject"].split(".")[-1]
    aux = r["auxiliary"]
    comparison = r["comparison"]
    if is_number(v=r["bound"]["value"]):
        value = r["bound"]["value"]
    elif r["bound"]["value"] == "t.b.d.":
        value = "t.b.d."
    else:
        value = r["bound"]["value"].split(".")[-1]
    unit = ""
    if r["bound"]["unit"]:
        unit = r["bound"]["unit"]

    return "{} {} {} {} {}".format(s, aux, comparison, value, unit)


@hookimpl
def function_spec_node(node: Node, graph: Graph) -> LineGen:
    if node.annotations.esl_info["sub_kind"] == "transformation":
        comp = node.annotations.esl_info["body"]["active"].split(".")[-1]
        aux = node.annotations.esl_info["body"]["auxiliary"]
        verb = node.annotations.esl_info["body"]["verb"]
        in_flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["input_variables"]
                        ]
                    )
                )
            )
        )
        prep = node.annotations.esl_info["body"]["preposition"]
        out_flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["output_variables"]
                        ]
                    )
                )
            )
        )
        line = "{} {} {} {} {} {}".format(comp, aux, verb, in_flows, prep, out_flows)

        # Bundle root variable dict
        brvdict = defaultdict(list)
        for flow in (
            node.annotations.esl_info["body"]["input_variables"]
            + node.annotations.esl_info["body"]["output_variables"]
        ):
            r = graph[flow].annotations.esl_info["bundle_root_name"]
            if r:
                brvdict[r].append(flow)

    elif node.annotations.esl_info["sub_kind"] == "goal":
        active = node.annotations.esl_info["body"]["active"].split(".")[-1]
        aux = node.annotations.esl_info["body"]["auxiliary"]
        verb = node.annotations.esl_info["body"]["verb"]
        flows = pm.hook.linguistic_enumeration(
            items=sorted(
                list(
                    set(
                        [
                            graph[flow].annotations.esl_info["bundle_root_name"]
                            if graph[flow].annotations.esl_info["bundle_root_name"]
                            else flow.split(".")[-1]
                            for flow in node.annotations.esl_info["body"]["variables"]
                        ]
                    )
                )
            )
        )
        prep = node.annotations.esl_info["body"]["preposition"]
        passive = node.annotations.esl_info["body"]["passive"].split(".")[-1]
        line = "{} {} {} {} {} {}".format(active, aux, verb, flows, prep, passive)

        # Bundle root varaible dict
        brvdict = defaultdict(list)
        for flow in node.annotations.esl_info["body"]["variables"]:
            r = graph[flow].annotations.esl_info["bundle_root_name"]
            if r:
                brvdict[r].append(flow)

    if node.annotations.esl_info["body"]["subclauses"]:
        yield lns.cap(line) + ", with subclauses:\n"
        yield from lns.unordered(
            [
                "\n".join(subclause_line(s=sc))
                for sc in node.annotations.esl_info["body"]["subclauses"]
            ]
        )
    else:
        yield lns.snt(line)

    if brvdict:
        yield from lns.bundle_clarification(brvdict=brvdict)


@hookimpl
def design_spec_node(node: Node, graph: Graph) -> LineGen:
    line = " or ".join([design_rule_line(r) for r in node.annotations.esl_info["body"]])
    if node.annotations.esl_info.get("sub_clauses", "node"):
        yield lns.cap(line) + ", with subclauses:\n"
        yield from lns.unordered(
            ["\n".join(subclause_line(s=sc)) for sc in node.annotations.esl_info["sub_clauses"]]
        )
    else:
        yield lns.snt(line)

    bvars = [
        v
        for v in lns.get_design_rule_line_vars(rules=node.annotations.esl_info["body"], g=graph)
        if v.annotations.esl_info["bundle_root_name"]
    ]

    if bvars:
        yield from lns.var_clarification(bvars=bvars)


@hookimpl
def behavior_spec_node(node: Node, graph: Graph) -> LineGen:
    clauses = []
    for case in node.annotations.esl_info["cases"]:
        yield lns.cap("case {}:\n".format(lns.emph(case["name"])))
        yield ""
        yield "when:\n"
        yield from lns.unordered(["\n".join(subclause_line(s=sc)) for sc in case["when_clauses"]])
        yield "then:\n"
        yield from lns.unordered(["\n".join(subclause_line(s=sc)) for sc in case["then_clauses"]])
        clauses += case["when_clauses"] + case["then_clauses"]

    default = node.annotations.esl_info.get("default")
    if default:
        yield lns.cap("case {}:\n".format(lns.emph("default")))
        yield ""
        yield "when no other case applies, then:\n"
        yield from lns.unordered(["\n".join(subclause_line(s=sc)) for sc in default])
        clauses += default

    bvars = [
        v
        for s in clauses
        for v in lns.get_design_rule_line_vars(rules=s["body"], g=graph)
        if v.annotations.esl_info["bundle_root_name"]
    ]

    if bvars:
        yield from lns.var_clarification(bvars=bvars)


@hookimpl
def need_node(node: Node, graph: Graph) -> LineGen:
    subject = node.annotations.esl_info["subject"]
    line = "{} {}".format(
        subject.split(".")[-1],
        node.annotations.esl_info["text"],
    )
    yield lns.snt(line)

    if graph[subject].kind == "variable":
        if graph[subject].annotations.esl_info["bundle_root_name"]:
            yield from lns.var_clarification(bvars=[graph[subject]])



"""Document module."""
import logging
from pathlib import Path
from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Union

import raesl.doc.lines as lns
import raesl.doc.rich as rich
import raesl.doc.sections as secs
from raesl import logger
from raesl.doc.locales import register_locale

# Compile module is excluded during docs generation.
try:
    from raesl.compile import cli
except ImportError:
    pass

TEMPLATES = Path(__file__).parent / "templates"


class Doc:
    """Output document.

    Arguments:
        paths: ESL input files.
        language: Output language.
        prologue: Markdown document to include as a prologue.
        epilogue: Markdown document to include as a conclusion.
        goal_section: Goal section toggle.
        transformation_section: Transformation section toggle.
        behavior_section: Behavior section toggle.
        design_section: Design section toggle.
        need_section: Need section toggle.
        relation_section: Need section toggle.
        var_table: Var table toggle.
        rich: Format to create rich output content in, defaults to 'tex'.
        rich_opts: Rich output generation options.

    Keyword arguments:
        Document metadata. See pandoc documentation.
    """

    # Format used when exporting to a Markdown file.
    markdown_file_format = "markdown+table_captions+multiline_tables+pipe_tables"

    # Format used in Doc generation.
    markdown_generated = "markdown+table_captions+multiline_tables+pipe_tables+grid_tables"

    def __init__(
        self,
        *paths: Union[str, Path],
        language: str = "en",
        prologue: Optional[Path] = None,
        epilogue: Optional[Path] = None,
        goal_section: [bool] = True,
        transformation_section: [bool] = True,
        behavior_section: [bool] = True,
        design_section: [bool] = True,
        need_section: [bool] = True,
        relation_section: [bool] = True,
        var_table: [bool] = True,
        rich: Optional[str] = "tex",
        rich_opts: Optional[Dict[str, Any]] = None,
        esl_paths: Optional[Union[List[str], List[Path]]] = None,
        **metadata,
    ):
        if esl_paths is not None:
            msg = " ".join(
                (
                    "The 'esl_paths' keyword argument will be deprecated.",
                    "Please use your file and directory paths as (any number of)",
                    "positional arguments to this function.",
                )
            )
            logger.warning(msg)
            paths = tuple(esl_paths)

        self.language = language
        register_locale(language)

        self.prologue = Path(prologue).read_text(encoding="utf-8") if prologue else None
        self.epilogue = Path(epilogue).read_text(encoding="utf-8") if epilogue else None

        self.goal_section = goal_section
        self.transformation_section = transformation_section
        self.behavior_section = behavior_section
        self.design_section = design_section
        self.need_section = need_section
        self.relation_section = relation_section
        self.var_table = var_table

        self.rich = rich
        self.rich_opts = rich_opts or dict()

        logger.debug("Parsing metadata {}...".format(metadata))
        self.metadata = {
            "book": True,
            "documentclass": "scrbook",
            "geometry": "margin=2.5cm",
            "graphics": True,
            "papersize": "a4",
            "titlepage": True,
            "titlepage-color": "1c3b6c",
            "titlepage-text-color": "ffffff",
            "titlepage-rule-color": "ffffff",
            "titlepage-background": (TEMPLATES / "background.pdf").as_posix(),
            "toc-own-page": True,
            "top-level-division": "chapter",
            "first-chapter": 1,
        }

        if metadata:
            self.metadata.update(metadata)

        self.metadata["lang"] = self.language

        logger.debug("Parsing ESL files...")

        self.diag_store, _, self.graph = cli.run(*paths)

        if self.diag_store.diagnostics:
            self.diag_store.dump()

        logger.debug("Parsing graph to a document...")
        self.pars: List[Par] = []
        self.parse_esl()

    def parse_esl(self):
        """Parse an ESL output Graph"""
        if self.prologue:
            self.pars = [Par([self.prologue])]

        g = self.graph

        comps = g["world"].children
        comps = sorted(comps, key=lambda x: x.name)
        depth = 0
        h = 1

        while comps:
            # Add chapter introduction.
            logger.debug("Processing nodes at level {}...".format(depth))
            self.pars.append(
                Par(
                    secs.node_decomp_level(
                        depth, comps, h=h, rich=self.rich, rich_opts=self.rich_opts
                    )
                )
            )
            # Add MDM figure in rich mode.
            if self.rich:
                logger.debug("Adding MDM image at level {}...".format(depth))
                self.pars.append(
                    Par(
                        rich.mdm(
                            graph=self.graph,
                            depth=depth,
                            rich=self.rich,
                            rich_opts=self.rich_opts,
                        )
                    )
                )

            # Add bundle information
            self.set_var_bundle_roots()

            for comp in comps:
                logger.debug("Adding sections for node {}...".format(comp.name))
                self.pars.append(Par(lns.lines(comp, h=h + 1)))
                self.pars.append(Par(secs.comp_node_props(comp, g, h + 2)))
                if self.goal_section:
                    self.pars.append(Par(secs.comp_node_goal_reqs(comp, g, h + 2)))
                    self.pars.append(Par(secs.comp_node_goal_cons(comp, g, h + 2)))

                if self.transformation_section:
                    self.pars.append(Par(secs.comp_node_transformation_reqs(comp, g, h + 2)))
                    self.pars.append(Par(secs.comp_node_transformation_cons(comp, g, h + 2)))

                if self.behavior_section:
                    self.pars.append(Par(secs.comp_node_behavior_reqs(comp, g, h + 2)))
                    self.pars.append(Par(secs.comp_node_behavior_cons(comp, g, h + 2)))

                if self.design_section:
                    self.pars.append(Par(secs.comp_node_design_reqs(comp, g, h + 2)))
                    self.pars.append(Par(secs.comp_node_design_cons(comp, g, h + 2)))

                if self.need_section:
                    self.pars.append(Par(secs.comp_node_needs(comp, g, h + 2)))

                if self.relation_section:
                    self.pars.append(Par(secs.comp_node_relations(comp, g, h + 2)))

                self.pars.append(Par(secs.comp_node_subcomps(comp, h + 2)))

            children = []
            for comp in comps:
                children.extend(comp.children)
            comps = children
            depth += 1

        # Check for any miscellaneous needs and design specifications that haven't
        # been printed yet.
        gn = [line for line in secs.global_needs(g, h + 1) if self.need_section]
        gdr = [line for line in secs.global_design_reqs(g, h + 1) if self.design_section]
        gdc = [line for line in secs.global_design_cons(g, h + 1) if self.design_section]
        if len(gn) > 0 or len(gdr) > 0 or len(gdc) > 0:
            self.pars.append(Par(secs.global_needs_and_designs(h)))
            self.pars.append(Par(gn))
            self.pars.append(Par(gdr))
            self.pars.append(Par(gdc))

        if self.epilogue:
            logger.debug("Adding epilogue...")
            self.pars.append(Par([self.epilogue]))

        # Appendices
        if self.var_table:
            logger.debug("Adding appendices...")
            self.pars.append(Par(["\\appendix{}", "\\appendixpage{}"]))
            self.pars.append(Par(secs.var_node_table(g, h=1)))

    @property
    def as_markdown(self) -> str:
        """Markdown representation of this document."""
        return "\n".join(self.yield_markdown())

    def yield_markdown(self) -> Generator[str, None, None]:
        """Yield markdown lines."""
        yield from self.yield_metadata()
        yield from self.yield_pars()

    def yield_metadata(self) -> Generator[str, None, None]:
        """Yield metadata lines."""
        yield "---"
        for key, value in self.metadata.items():
            yield "{}: {}".format(key, value)
        yield "---"

    def yield_pars(self) -> Generator[str, None, None]:
        """Yield all paragraph texts."""
        for par in self.pars:
            yield par.md

    def save(
        self,
        path: Union[Path, str],
        to: Optional[str] = None,
        pandoc_args: List[str] = [
            "--standalone",
            "--number-sections",
            "--toc",
            "--listings",
            "--self-contained",
        ],
        filters: List[str] = ["pandoc-fignos"],
    ):
        """Save document as a file.

        Arguments:
            path: Path to save to.
            to: Optional format to save to. Normally derived from path.
            pandoc_args: Additional arguments for pandoc conversion tool.
            filters: Pandoc filters to use.
        """
        import pypandoc

        path = Path(path)
        if to is None:
            to = path.suffix.lstrip(".")
            if to == "html":
                to == "html5"
        pandoc_args = pandoc_args.copy()
        filters = filters.copy()

        # Fix LaTeX incompatibilities.
        if (to == "tex" or to == "pdf") and self.rich == "md":
            self.rich = "tex"
            self.parse_esl()

        if logger.level <= logging.DEBUG and to != "md":
            logger.debug("Saving Markdown file for debugging purposes...")
            try:
                pypandoc.convert_text(
                    self.as_markdown,
                    self.markdown_file_format,
                    format=self.markdown_generated,
                    outputfile=str(path.with_suffix(".md")),
                    encoding="utf-8",
                    extra_args=pandoc_args,
                    filters=filters,
                )
            except RuntimeError as exc:
                logger.error(str(exc))

        if to == "pdf" or to == "latex":
            template = (TEMPLATES / "eisvogel.latex").resolve()
            pandoc_args.extend(["--template", str(template)])

        if logger.level <= logging.DEBUG and to != "latex":
            try:
                logger.debug("Saving LaTeX file for debugging purposes...")
                pypandoc.convert_text(
                    self.as_markdown,
                    "latex",
                    format=self.markdown_generated,
                    outputfile=str(path.with_suffix(".tex")),
                    encoding="utf-8",
                    extra_args=pandoc_args,
                    filters=filters,
                )
            except RuntimeError as exc:
                logger.error(str(exc))

        logger.debug("Converting to {}...".format(to.upper()))

        # Convert and save.
        try:
            pypandoc.convert_text(
                self.as_markdown,
                to,
                format=self.markdown_generated,
                outputfile=str(path),
                encoding="utf-8",
                extra_args=pandoc_args,
                filters=filters,
            )
        except RuntimeError as exc:
            logger.error(str(exc))
            raise exc

    def get_bundle_name_parts(self) -> Set[str]:
        """Get set of all name parts of bundles used within the spec

        Returns:
            Set of all name parts of bundles.
        """
        vnames = [n.name for n in self.graph.get_nodes_by_kind("variable")]
        cnames = [n.name for n in self.graph.get_nodes_by_kind("component")]

        cname_parts = set()
        for name in cnames:
            cname_parts.update(set(name.split(".")))

        vname_parts = set()
        for name in vnames:
            vname_parts.update(set(name.split(".")[:-1]))

        return vname_parts - cname_parts

    def set_var_bundle_roots(self):
        """Set the bundle root of variables if they originate from a bundle."""
        bnps = self.get_bundle_name_parts()
        for v in self.graph.get_nodes_by_kind("variable"):
            v.annotations.esl_info["bundle_root_name"] = get_bundle_root(
                vname=v.name, bundle_name_parts=bnps
            )


class Par:
    """Paragraph.

    Arguments:
        lines: Lines of this paragraph.
        obj: Object of this paragraph.

    Attributes:
        md: Markdown representation.
    """

    def __init__(self, lines: Iterable[str], obj: Optional[Any] = None):
        self.md = "\n".join(lines) + "\n"
        self.md = self.md.replace("_", " ").replace(" ->", " &rarr;")
        self.md = self.md.replace("-underscore-", "_")
        self.obj = obj


def get_bundle_root(vname: str, bundle_name_parts: Set) -> Union[None, str]:
    """Check if variable originates from a bundle and return the name
    of the root of the bundle.

    Arguments:
        vname: Name of the variable to check.
        bundle_name_parts: Set of strings the comprise bundle names.

    Returns:
       None or Name of the root of the bundle from which the variable originates.
    """
    bundle_root = None
    name_parts = vname.split(".")

    if name_parts[-2] not in bundle_name_parts:
        # Variable did not originate from a bundle:
        return bundle_root

    for idx in range(2, len(name_parts) + 1):
        if name_parts[-idx] in bundle_name_parts:
            bundle_root = name_parts[-idx]
        else:
            # Location part of varaible name has been entered
            break

    return bundle_root



"""Class for constructing types."""
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from raesl.compile import diagnostics, scanner
from raesl.compile.ast import exprs, types
from raesl.compile.typechecking.orderer import Orderer

if TYPE_CHECKING:
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder
    from raesl.compile.typechecking.orderer import OrdererEntry
    from raesl.types import Location

STANDARD_TYPE_NAMES = ("real", "integer", "string", "boolean")


class TempTypeDef:
    """Temporary storage of a type definition."""

    def __init__(
        self,
        type_name: "Token",
        parent_name: Optional["Token"],
        enum_spec: Optional[List[exprs.Value]],
        unit_spec: Optional[List["Token"]],
        ival_spec: Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]],
        cons_spec: Optional[exprs.Value],
    ):
        self.type_name = type_name
        self.parent_name = parent_name
        self.enum_spec = enum_spec
        self.unit_spec = unit_spec
        self.ival_spec = ival_spec
        self.cons_spec = cons_spec


class TempFieldDef:
    """Temporary storage of a field in a bundle."""

    def __init__(
        self,
        field_name: "Token",
        type_name: Optional["Token"],
    ):
        self.field_name = field_name
        self.type_name = type_name


class TempBundleDef:
    """Temporary storage of a bundle definition."""

    def __init__(self, bundle_name: "Token"):
        self.bundle_name = bundle_name
        self.bundle_fields: List[TempFieldDef] = []


class TypeBuilder:
    """Builder to construct types of the specification.

    Arguments:
        ast_builder: AST builder instance.

    Attributes:
        diag_store: Storage for found diagnostics.
        type_defs: Found type definitions in the specification, temporarily stored until
            all type information is available.
        bundle_defs: Found bundle definitions in the specification, temporarily stored
            until all type information is available.
        current_bundle: If not None, reference to the last opened bundle definition
            for adding additional fields to it.
    """

    def __init__(self, ast_builder: "AstBuilder"):
        # Make the builder problem store available locally.
        self.diag_store = ast_builder.diag_store

        # Setup type data storage.
        self.type_defs: List[TempTypeDef] = []
        self.bundle_defs: List[TempBundleDef] = []
        self.current_bundle: Optional[TempBundleDef] = None
        self.types_with_error: Optional[
            Dict[str, Token]
        ] = None  # typedefs that were not added setup during type construction.

        ast_builder.register_new_section(self)

    def notify_new_section(self, _new_top_section: bool):
        """Notification that type additions have finished."""
        self.current_bundle = None

    @staticmethod
    def get_entry_location(entry: "OrdererEntry") -> "Location":
        """Retrieve position information from a type entry.

        Arguments:
            entry: Entry to use.

        Returns:
            Position information of the entry.
        """
        if isinstance(entry.data, TempTypeDef):
            return entry.data.type_name.get_location()
        else:
            assert isinstance(entry.data, TempBundleDef)
            return entry.data.bundle_name.get_location()

    @staticmethod
    def add_standard_types(resolved_types: Dict[str, types.TypeDef]):
        """Add all standard types to the resolved types."""
        for name in STANDARD_TYPE_NAMES:
            name_tok = scanner.Token("NAME", name, "ESL-compiler", 0, 0, 0)
            standard_type = types.ElementaryType(None, [], None)
            typedef = types.TypeDef(name_tok, standard_type)

            assert name not in resolved_types
            resolved_types[name] = typedef

    def finish(self, spec):
        """Check the collected types and bundles, report any errors, and add them
        to the specification.

        Arguments:
            spec: Specification to extend with the found types.
        """
        self.current_bundle = None  # Likely not needed, but it's safe.

        orderer = Orderer()

        # Add type definitions.
        for tdef in self.type_defs:
            if tdef.parent_name is None:
                needs = []
            else:
                needs = [tdef.parent_name.tok_text]

            # Check for duplicate use of the name as a type.
            entry = orderer.find_entry(tdef.type_name.tok_text)
            if entry is not None:
                locs = [
                    tdef.type_name.get_location(),
                    TypeBuilder.get_entry_location(entry),
                ]
                self.diag_store.add(
                    diagnostics.E200(
                        tdef.type_name.tok_text,
                        "type definition",
                        location=locs[0],
                        dupes=locs,
                    )
                )
                continue

            orderer.add_dependency(tdef.type_name.tok_text, needs, tdef)

        # Add bundle definitions
        for bdef in self.bundle_defs:
            needs = set()
            field_names = {}
            for bfield in bdef.bundle_fields:
                # Verify unique field names.
                field_name = bfield.field_name.tok_text
                if field_name in field_names:
                    locs = [
                        field_names[field_name].get_location(),
                        bfield.field_name.get_location(),
                    ]
                    self.diag_store.add(
                        diagnostics.E200(
                            field_name,
                            "Bundle field name",
                            location=locs[0],
                            dupes=locs,
                        )
                    )
                    # Continue anyway

                if bfield.type_name is not None:
                    needs.add(bfield.type_name.tok_text)

            # Check for duplicate use of the name as a type.
            entry = orderer.find_entry(bdef.bundle_name.tok_text)
            if entry is not None:
                locs = [
                    bdef.bundle_name.get_location(),
                    TypeBuilder.get_entry_location(entry),
                ]
                # Type name is used for more than one type
                self.diag_store.add(
                    diagnostics.E200(
                        bdef.bundle_name.tok_text, "type", location=locs[0], dupes=locs
                    )
                )
                continue

            orderer.add_dependency(bdef.bundle_name.tok_text, list(needs), bdef)

        # Let the orderer decide on an order of processing, and process the result.
        resolved, cycle = orderer.resolve()

        self.types_with_error = (
            {}
        )  # (str -> Token) Types that are defined in the spec but something was wrong.
        resolved_types = {}  # Dict of name-string to types.TypeDef.
        # Make standard types generally available for the entire specification.
        TypeBuilder.add_standard_types(resolved_types)

        for entry in resolved:
            if entry.data is None:
                # Entry was created by the orderer, we should run into it again and fail
                # to find it.
                continue

            if isinstance(entry.data, TempTypeDef):
                self._make_typedef(entry.data, resolved_types)
            else:
                assert isinstance(entry.data, TempBundleDef)
                self._make_bundledef(entry.data, resolved_types)

        self.types_with_error = None

        if cycle:
            locs = [TypeBuilder.get_entry_location(entry) for entry in cycle]
            name = cycle[0].name
            self.diag_store.add(
                diagnostics.E204(name, "type definition", location=locs[0], cycle=locs)
            )

        # Store output
        spec.types = resolved_types

    def _construct_type(
        self,
        parent_name: Optional["Token"],
        enum_spec: Optional[List[exprs.Value]],
        unit_spec: Optional[List["Token"]],
        ival_spec: Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]],
        cons_spec: Optional[exprs.Value],
        prefer_parent: bool,
        resolved_types: Dict[str, types.TypeDef],
    ) -> Optional[types.BaseType]:
        """Construct a new type from the given parameters.

        Arguments:
            prefer_parent: If possible, don't make a new type.

        Returns:
            The parent type if nothing changed, the constructed ElementaryType if
                something was extended, or None if construction failed.
        """
        # Find parent type, and verify we can extend it if needed.
        parent_type: Optional[types.BaseType]
        if parent_name is None:
            parent_type = None
        else:
            # Type should be available.
            parent_typedef = resolved_types.get(parent_name.tok_text)
            if parent_typedef is None:
                # Not available, perhaps an error happened with that type?
                assert self.types_with_error is not None
                parent_def_tok = self.types_with_error.get(parent_name.tok_text)

                if parent_def_tok:
                    # Parent was defined but failed with an error.
                    self.diag_store.add(
                        diagnostics.E214(
                            parent_name.tok_text,
                            location=parent_name.get_location(),
                            def_location=parent_def_tok.get_location(),
                        )
                    )
                else:
                    # Parent was never defined at all.
                    self.diag_store.add(
                        diagnostics.E203(
                            "type",
                            name=parent_name.tok_text,
                            location=parent_name.get_location(),
                        )
                    )
                return None

            parent_type = parent_typedef.type
            is_extended = (
                (enum_spec is not None)
                or (unit_spec is not None)
                or (ival_spec is not None)
                or (cons_spec is not None)
            )

            if not isinstance(parent_type, types.ElementaryType):
                # Parent is a compound, type cannot be extended.
                if is_extended:
                    self.diag_store.add(
                        diagnostics.E215(parent_name.tok_text, location=parent_name.get_location())
                    )
                    return None

                return parent_type

            # Parent is an elementary type.
            if not is_extended and prefer_parent:
                # No extension and a parent type is preferred.
                return parent_type

        # Check the data of the new type.
        #
        # Process units.
        available_units: Dict[str, "Token"] = {}
        if parent_type:
            available_units = dict((tok.tok_text, tok) for tok in parent_type.units)
        else:
            available_units = {}

        type_units = []
        if unit_spec:
            for unit in unit_spec:
                unit_text: str = unit.tok_text
                # Check for not having '[xyz]' or '-'.
                if unit_text.startswith("[") and unit_text.endswith("]"):
                    self.diag_store.add(
                        diagnostics.E216(unit.tok_text, location=unit.get_location())
                    )
                    return None
                if unit_text == "-":
                    self.diag_store.add(diagnostics.E217(location=unit.get_location()))
                    return None

                if unit_text in available_units:
                    self.diag_store.add(
                        diagnostics.W200(unit.tok_text, "unit", location=unit.get_location())
                    )
                    continue  # Silently discard duplicates.

                type_units.append(unit)
                available_units[unit_text] = unit

        # Build intervals
        type_intervals: Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]]
        type_intervals = None

        if ival_spec:
            type_intervals = ival_spec

        if enum_spec:
            assert not type_intervals
            type_intervals = [(val, val) for val in enum_spec]

        if cons_spec:
            assert not type_intervals
            type_intervals = [(cons_spec, cons_spec)]

        if not type_intervals:
            type_intervals = []

        if parent_type and prefer_parent and not type_units and not type_intervals:
            # An extension was specified but nothing changed, keep the parent if
            # preferred.
            return parent_type

        # Verify units of the type limits.
        for low_val, high_val in type_intervals:
            self.check_unit(low_val, available_units)
            self.check_unit(high_val, available_units)

        assert type_intervals is not None
        return types.ElementaryType(parent_type, type_units, type_intervals)

    def _check_override_standard_type(self, type_name: "Token") -> bool:
        """Check whether the new type tries to override a standard type. Give an error
        if it does.

        Returns:
            True if the new type is a standard type, else False.
        """
        if type_name.tok_text in STANDARD_TYPE_NAMES:
            self.diag_store.add(
                diagnostics.E218(type_name.tok_text, location=type_name.get_location)
            )
            return True
        return False

    def _make_typedef(self, tdef: TempTypeDef, resolved_types: Dict[str, types.TypeDef]):
        """Add the type definition to the resolved collection. It is assumed the
        dependencies are already available (as ensured by the Orderer).
        """
        if self._check_override_standard_type(tdef.type_name):
            return  # Something failed and was reported.

        new_type = self._construct_type(
            tdef.parent_name,
            tdef.enum_spec,
            tdef.unit_spec,
            tdef.ival_spec,
            tdef.cons_spec,
            False,
            resolved_types,
        )
        if new_type is None:
            assert self.types_with_error is not None
            self.types_with_error[tdef.type_name.tok_text] = tdef.type_name
            return  # Something failed and was reported.

        typedef = types.TypeDef(tdef.type_name, new_type)
        assert typedef.name.tok_text not in resolved_types
        resolved_types[typedef.name.tok_text] = typedef

    def _make_bundledef(self, bdef: TempBundleDef, resolved_types: Dict[str, types.TypeDef]):
        """Add the bundle definition to the resolved collection. It is assumed the
        dependencies are already available (as ensured by the Orderer).
        """
        if self._check_override_standard_type(bdef.bundle_name):
            return

        fields = []
        field_names: Dict[str, "Token"] = {}  # For duplicate field name detection.
        failed = False
        for field in bdef.bundle_fields:
            new_type = self._construct_type(
                field.type_name,
                None,
                None,
                None,
                None,
                True,
                resolved_types,
            )
            if not new_type:
                assert self.types_with_error is not None
                self.types_with_error[bdef.bundle_name.tok_text] = bdef.bundle_name
                failed = True
                continue  # But continue looking for more errors.

            if field.field_name.tok_text in field_names:
                # Duplicate field, discard. Above it was already reported.
                continue

            field_names[field.field_name.tok_text] = field.field_name
            fields.append(types.CompoundField(field.field_name, new_type))

        if failed:
            # Some error was detected, bail out.
            return

        new_bundle = types.Compound(fields)
        typedef = types.TypeDef(bdef.bundle_name, new_bundle)
        assert typedef.name.tok_text not in resolved_types
        resolved_types[typedef.name.tok_text] = typedef

    def add_typedef(
        self,
        type_name: "Token",
        parent_name: Optional["Token"],
        enum_spec: Optional[List[exprs.Value]],
        unit_spec: Optional[List["Token"]],
        ival_spec: Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]],
        cons_spec: Optional[exprs.Value],
    ):
        """The parser found a new type definition entry, store it."""
        temp_tdef = TempTypeDef(type_name, parent_name, enum_spec, unit_spec, ival_spec, cons_spec)
        self.type_defs.append(temp_tdef)
        self.current_bundle = None

    def new_bundle_type(self, bundle_name: "Token"):
        """The parser found a new bundle in the source code, create it."""
        temp_bdef = TempBundleDef(bundle_name)
        self.bundle_defs.append(temp_bdef)
        self.current_bundle = temp_bdef

    def add_bundle_field(
        self,
        field_name: "Token",
        type_name: Optional["Token"],
    ):
        """A new field in the current bundle has been found by the parser, add it."""
        temp_fdef = TempFieldDef(field_name, type_name)
        assert self.current_bundle, "Trying to add a bundle field outside a type section."
        self.current_bundle.bundle_fields.append(temp_fdef)

    def check_unit(self, value: Optional[exprs.Value], available_units: Dict[str, "Token"]) -> None:
        """Check whether the unit possibly specified in 'value' is available for use.
        Report an error if it is not available.
        """
        if value is None or value.unit is None:
            return

        units = value.get_units()
        if units is None:
            return

        if units.intersection(available_units):
            return  # Value uses a known unit.

        unit_text = value.unit.tok_text
        if unit_text.startswith("[") and unit_text.endswith("]"):
            unit_text = unit_text[1:-1]
        self.diag_store.add(diagnostics.E219(unit_text, location=value.unit.get_location()))
